from typing import Optional
import pandas as pd
import numpy as np
from plotly.offline import init_notebook_mode, iplot
import plotly.graph_objects as go
from plotly.subplots import make_subplots


def select_numeric(x: pd.DataFrame) -> pd.DataFrame:
    x = x.select_dtypes(include=[np.number], exclude=[bool])

    # print(x)
    return x


def select_categorical_text(x: pd.DataFrame) -> pd.DataFrame:
    x = x.select_dtypes(exclude=[np.number])
    """
    Select non numeric columns from a Dataframe

    Returns:
        pd.DataFrame: DataFrame with just categorical data
    """
    # print(x)
    return x


def calculate_distribution(
    df: pd.DataFrame, variable: str, target_name: str, target_value_described: str
) -> pd.DataFrame:

    df = get_variable_and_target(df, variable, target_name)

    dummy_names = [target_name + "_" +
                   str(dummy) for dummy in df[target_name].unique()]
    target_value_described = target_name + "_" + target_value_described
    df = pd.get_dummies(data=df, columns=[target_name])
    df = df.groupby(by=variable)[dummy_names].sum()
    df["Total"] = df.sum(axis=1).astype(int)
    df = df.astype(int).reset_index()
    df["Percentage"] = (df[target_value_described] / df["Total"]) * 100

    # x = (
    #     df[df[target_name] == df[target_name].unique()[-1]]
    #     .groupby(variable)[target_name]
    #     .count()
    # )
    # y = (
    #     df[df[target_name] == df[target_name].unique()[0]]
    #     .groupby(variable)[target_name]
    #     .count()
    # )

    # vv = pd.DataFrame({variable: df[variable].unique()})

    # vv = pd.merge(vv, x, how="left", right_index=True, left_on=variable)
    # vv = pd.merge(
    #     vv, y, how="left", right_index=True, left_on=variable, suffixes=("", "_not")
    # )
    # vv.fillna(0, inplace=True)
    # vv["Total"] = vv[target_name] + vv[target_name + "_not"]
    # vv["Percentage"] = (vv[target_name] / vv["Total"]) * 100

    return df


def sample_and_get_distribution(
    df: pd.DataFrame,
    variable: str,
    target_name: str,
    target_value_described: str,
    sample_size: int,
    random_state: Optional[int] = None,
) -> pd.DataFrame:
    df = get_variable_and_target(df, variable, target_name=target_name)
    dummy_names = [target_name + "_" +
                   str(dummy) for dummy in df[target_name].unique()]
    target_value_described = target_name + "_" + target_value_described
    df = pd.get_dummies(data=df, columns=[target_name])
    df = df.groupby(by=variable)[dummy_names].sum()

    if random_state:
        sample = df.sample(n=sample_size, random_state=random_state)
    else:
        sample = df.sample(n=sample_size)

    resto = df.drop(sample.index)
    resto = resto.sum()
    resto = pd.DataFrame(resto).T
    resto.index = ["All others"]

    df = pd.concat([sample, resto])
    df.index.name = variable

    df["Total"] = df.sum(axis=1).astype(int)
    df = df.astype(int).reset_index()
    df["Percentage"] = (df[target_value_described] / df["Total"]) * 100
    return df


def calculate_bins(
    df: pd.DataFrame,
    variable: str,
    target_name: str,
    target_value_described: str,
    nbins: int,
):
    df = get_variable_and_target(df, variable, target_name)
    dummy_names = [target_name + "_" +
                   str(dummy) for dummy in df[target_name].unique()]
    dummy_dct = {dummy_name: "sum" for dummy_name in dummy_names}
    target_value_described = target_name + "_" + target_value_described
    df = pd.get_dummies(data=df, columns=[target_name])
    df = df[df[variable].notna()]  # we ommit the na
    counts, bins = np.histogram(df[variable], bins=nbins)
    # bins = 0.5 * (bins[:-1] + bins[1:])
    df.loc[:, "Belong"] = np.digitize(df[variable], bins)
    df = df.groupby(by="Belong").agg({**{variable: "median"}, **dummy_dct})
    # df = df.reindex(list(range(df.index.min(), df.index.max() + 1)), fill_value=0)
    df["Total"] = df[dummy_names].sum(axis=1)
    df["Percentage"] = (df[target_value_described] / df["Total"]) * 100
    # df["Percentage"].fillna(0, inplace=True)
    return df, counts, bins


def get_variable_and_target(
    df: pd.DataFrame, variable_name: str, target_name: str
) -> pd.DataFrame:
    return df[[variable_name, target_name]]


def plot_variable(
    df: pd.DataFrame,
    variable_name: str,
    target_name: str,
    target_value_described: str,
    export: bool = False,
) -> None:

    fig = make_subplots(specs=[[{"secondary_y": True}]])
    fig.add_trace(
        go.Bar(x=df[variable_name], y=df["Total"], marker_color="#1f77b4"),
        secondary_y=False,
    )
    fig.add_trace(
        go.Scatter(
            x=df[variable_name],
            y=df["Percentage"],
            mode="markers",
            marker={"size": 20},
            marker_color="#fd3216",
        ),
        secondary_y=True,
    )

    fig.update_xaxes(title_text=f"{variable_name}", type="category")
    fig.update_yaxes(
        title_text="Number of rows",
        secondary_y=False,
        title_font=dict(color="#1f77b4"),
        color="#1f77b4",
    )
    fig.update_yaxes(
        title_text=f"Percentage of {target_name} = {target_value_described}",
        secondary_y=True,
        title_font=dict(color="#fd3216"),
        color="#fd3216",
    )
    fig.update_layout(showlegend=False)
    if export:
        fig.write_html(f"{variable_name+'_'+target_value_described}.html")

    if not export:
        # initiate notebook for offline plot
        init_notebook_mode(connected=True)
        iplot(fig)


def plot_numerical_variable(
    df: pd.DataFrame,
    grouped_bins: pd.DataFrame,
    counts,
    bins,
    variable_name: str,
    target_name: str,
    target_value_described: str,
    nbins: int,
    nbins_round_2: Optional[dict] = None,
    export: bool = False,
) -> None:

    size = np.ediff1d(bins)[0]
    min_hist = np.min(bins)
    max_hist = np.max(bins)

    fig = make_subplots(specs=[[{"secondary_y": True}]])
    # hist = px.histogram(df, x=variable_name, nbins=nbins)
    fig.add_trace(
        go.Histogram(
            x=df[variable_name],
            xbins={"start": min_hist, "size": size, "end": max_hist + 1},
            marker_color="#1f77b4",
        ),
        secondary_y=False,
    )
    # nbins = (
    # px.histogram(
    # df, x=variable_name, nbins=nbins, color_discrete_sequence=["#1f77b4"]
    # )
    # .full_figure_for_development()
    # .data[0]
    # .xbins["size"]
    # )

    # grouped_bins = calculate_bins(
    # df, variable_name, target_name, target_value_described, nbins
    # )

    # fig.add_trace(
    # go.Bar(x=bins, y=counts, marker_color="#1f77b4"), secondary_y=False,
    # )
    if nbins_round_2:
        if variable_name in nbins_round_2.keys():
            grouped_bins = calculate_bins(
                df,
                variable_name,
                target_name,
                target_value_described,
                nbins_round_2[variable_name] - 1,  # forsousky
            )

    fig.add_trace(
        go.Scatter(
            x=grouped_bins[variable_name],
            # x=bins,
            y=grouped_bins["Percentage"],
            mode="markers+lines",
            marker={"size": 20},
            marker_color="#fd3216",
        ),
        secondary_y=True,
    )

    fig.update_xaxes(title_text=f"{variable_name}")
    fig.update_yaxes(
        title_text="Number of rows",
        secondary_y=False,
        title_font=dict(color="#1f77b4"),
        color="#1f77b4",
    )
    fig.update_yaxes(
        title_text=f"Percentage of {target_name} = {target_value_described}",
        secondary_y=True,
        title_font=dict(color="#fd3216"),
        color="#fd3216",
    )
    fig.update_layout(showlegend=False)
    if export:
        fig.write_html(f"{variable_name+'_'+target_value_described}.html")
    if not export:
        # initiate notebook for offline plot
        init_notebook_mode(connected=True)
        iplot(fig)



from typing import List, Literal, Optional, Union
import pandas as pd
from .utils import (
    calculate_bins,
    get_variable_and_target,
    plot_numerical_variable,
    sample_and_get_distribution,
    select_categorical_text,
    select_numeric,
    plot_variable,
    calculate_distribution,
)


class targetDescribe:
    def __init__(
        self,
        data: pd.DataFrame,
        target: Union[pd.Series, str],
        problem: Literal["binary_classification", "regression"],
        max_categories: int = 30,
        target_described: Optional[str] = None,
        nbins: int = 15,
    ) -> None:
        __target_in_df = False

        if isinstance(target, str):
            try:
                self._target_name = target
                self.target = data[target].copy()
                __target_in_df = True
            except KeyError:
                raise KeyError(f"{target} not in DataFrame")
        else:
            self._target_name = target.name
            self.target = target.copy()

        if target_described:
            self.target_value_described = target_described
        else:
            self.target_value_described = str(self.target.unique()[-1])
        self.nbins = nbins
        self.max_categories = max_categories
        self.data = data.copy()
        self.problem = problem

        self.split_variables()
        self.numeric_variables = self._append_target(
            variables=self.numeric_variables, target_in_df=__target_in_df
        )
        self.categorical_variables = self._append_target(
            variables=self.categorical_variables, target_in_df=__target_in_df
        )

    def split_variables(self) -> None:
        self.numeric_variables = select_numeric(self.data)
        self.categorical_variables = select_categorical_text(self.data)

    def _append_target(
        self, variables: pd.DataFrame, target_in_df: bool
    ) -> pd.DataFrame:

        if target_in_df:
            if self.target.name in list(variables.columns):
                return variables
            else:
                return pd.concat([variables, self.target], axis=1)
        else:
            return pd.concat([variables, self.target], axis=1)

    def all_associations(
        self,
        target_value_described: Optional[str] = None,
        export: bool = False,
        max_categories: Optional[int] = None,
        nbins: Optional[int] = None,
        random_state: Optional[int] = None,
        nbins_round_2: Optional[dict] = None,
    ):
        if nbins:
            self.nbins = nbins
        if max_categories:
            self.max_categories = max_categories

        if target_value_described:
            self.target_value_described = target_value_described

        if self.problem == "binary_classification":

            if self.target.nunique() != 2:
                raise ("Not binary target")

            numeric_columns = [
                name
                for name in list(self.numeric_variables.columns)
                if name not in [self._target_name]
            ]
            categorical_columns = [
                name
                for name in list(self.categorical_variables.columns)
                if name not in [self._target_name]
            ]

            for nombre in numeric_columns:
                if self.numeric_variables[nombre].dtype in ["int64", "int32", "int16"]:
                    num_categorias = self.numeric_variables[nombre].nunique()

                    if num_categorias <= self.max_categories:
                        proporcion = calculate_distribution(
                            df=self.numeric_variables,
                            variable=nombre,
                            target_name=self._target_name,
                            target_value_described=self.target_value_described,
                        )

                        plot_variable(
                            df=proporcion,
                            variable_name=nombre,
                            target_name=self._target_name,
                            target_value_described=self.target_value_described,
                            export=export,
                        )
                    else:

                        proporcion = sample_and_get_distribution(
                            df=self.numeric_variables,
                            variable=nombre,
                            target_name=self._target_name,
                            target_value_described=self.target_value_described,
                            sample_size=self.max_categories,
                            random_state=random_state,
                        )

                        plot_variable(
                            df=proporcion,
                            variable_name=nombre,
                            target_name=self._target_name,
                            target_value_described=self.target_value_described,
                            export=export,
                        )
                elif self.numeric_variables[nombre].dtype in ["float64", "float32"]:
                    proporcion, counts, bins = calculate_bins(
                        df=self.numeric_variables,
                        variable=nombre,
                        target_name=self._target_name,
                        target_value_described=self.target_value_described,
                        nbins=self.nbins,
                    )

                    plot_numerical_variable(
                        get_variable_and_target(
                            self.numeric_variables, nombre, self._target_name
                        ),
                        proporcion,
                        counts,
                        bins,
                        variable_name=nombre,
                        target_name=self._target_name,
                        target_value_described=self.target_value_described,
                        nbins=self.nbins,
                        nbins_round_2=nbins_round_2,
                        export=export,
                    )
            for nombre in categorical_columns:
                if self.categorical_variables[nombre].dtype == "object":
                    num_categorias = self.categorical_variables[nombre].nunique(
                    )
                    if num_categorias <= self.max_categories:

                        proporcion = calculate_distribution(
                            df=self.categorical_variables,
                            variable=nombre,
                            target_name=self._target_name,
                            target_value_described=self.target_value_described,
                        )

                        plot_variable(
                            df=proporcion,
                            variable_name=nombre,
                            target_name=self._target_name,
                            target_value_described=self.target_value_described,
                            export=export,
                        )

                    else:

                        proporcion = sample_and_get_distribution(
                            df=self.categorical_variables,
                            variable=nombre,
                            target_name=self._target_name,
                            target_value_described=self.target_value_described,
                            sample_size=self.max_categories,
                            random_state=random_state,
                        )

                        plot_variable(
                            df=proporcion,
                            variable_name=nombre,
                            target_name=self._target_name,
                            target_value_described=self.target_value_described,
                            export=export,
                        )

    def describe_some(self, columns: List[str], target_value_described: Optional[str] = None, export: bool = False, max_categories: Optional[int] = None, nbins: Optional[int] = None, random_state: Optional[int] = None, nbins_round_2: Optional[dict] = None):
        if nbins:
            self.nbins = nbins
        if max_categories:
            self.max_categories = max_categories

        if target_value_described:
            self.target_value_described = target_value_described

        if self.problem == "binary_classification":
            if self.target.nunique() != 2:
                raise ("Not binary target")

            for nombre in columns:
                if nombre in self.numeric_variables.columns:

                    if self.numeric_variables[nombre].dtype in ["int64", "int32", "int16"]:
                        num_categorias = self.numeric_variables[nombre].nunique(
                        )

                        if num_categorias <= self.max_categories:
                            proporcion = calculate_distribution(
                                df=self.numeric_variables,
                                variable=nombre,
                                target_name=self._target_name,
                                target_value_described=self.target_value_described,
                            )

                            plot_variable(
                                df=proporcion,
                                variable_name=nombre,
                                target_name=self._target_name,
                                target_value_described=self.target_value_described,
                                export=export,
                            )
                        else:

                            proporcion = sample_and_get_distribution(
                                df=self.numeric_variables,
                                variable=nombre,
                                target_name=self._target_name,
                                target_value_described=self.target_value_described,
                                sample_size=self.max_categories,
                                random_state=random_state,
                            )

                            plot_variable(
                                df=proporcion,
                                variable_name=nombre,
                                target_name=self._target_name,
                                target_value_described=self.target_value_described,
                                export=export,
                            )
                    elif self.numeric_variables[nombre].dtype in ["float64", "float32"]:
                        proporcion, counts, bins = calculate_bins(
                            df=self.numeric_variables,
                            variable=nombre,
                            target_name=self._target_name,
                            target_value_described=self.target_value_described,
                            nbins=self.nbins,
                        )

                        plot_numerical_variable(
                            get_variable_and_target(
                                self.numeric_variables, nombre, self._target_name
                            ),
                            proporcion,
                            counts,
                            bins,
                            variable_name=nombre,
                            target_name=self._target_name,
                            target_value_described=self.target_value_described,
                            nbins=self.nbins,
                            nbins_round_2=nbins_round_2,
                            export=export,
                        )

                elif nombre in self.categorical_variables.columns:

                    if self.categorical_variables[nombre].dtype == "object":
                        num_categorias = self.categorical_variables[nombre].nunique(
                        )
                        if num_categorias <= self.max_categories:

                            proporcion = calculate_distribution(
                                df=self.categorical_variables,
                                variable=nombre,
                                target_name=self._target_name,
                                target_value_described=self.target_value_described,
                            )

                            plot_variable(
                                df=proporcion,
                                variable_name=nombre,
                                target_name=self._target_name,
                                target_value_described=self.target_value_described,
                                export=export,
                            )

                        else:

                            proporcion = sample_and_get_distribution(
                                df=self.categorical_variables,
                                variable=nombre,
                                target_name=self._target_name,
                                target_value_described=self.target_value_described,
                                sample_size=self.max_categories,
                                random_state=random_state,
                            )

                            plot_variable(
                                df=proporcion,
                                variable_name=nombre,
                                target_name=self._target_name,
                                target_value_described=self.target_value_described,
                                export=export,
                            )

                else:
                    print(f"{nombre} no esta en el dataframe cainal")


if __name__ == "__main__":

    df = pd.read_csv(
        "https://raw.githubusercontent.com/datasciencedojo/datasets/master/titanic.csv"
    )
    hola = df["Survived"].astype(str)
    df.drop(axis=1, labels="Survived", inplace=True)
    # print(df.columns)
    # a = targetDescribe(data=df, target="Survived", problem="binary_classification")

    b = targetDescribe(data=df, target=hola, problem="binary_classification")

    b.all_associations(
        max_categories=10, export=True,
    )
    # b.all_associations(export=True, target_value_described="0")



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent
# The text of the README file
README = (HERE / "README.rst").read_text()
# This call to setup() does all the work
setup(
    name="target_describe",
    version="0.0.2",
    description="Generate viz for your variables with your target for ML",
    long_description=README,
    url="https://github.com/DanielR59/target_description",
    author="Daniel Rosas",
    author_email="daniel_pumas_59@hotmail.com",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    install_requires=open("requirements.txt").readlines(),
)
